import json
import logging
import os
import re
import unittest
from pprint import pformat

from httmock import HTTMock, urlmatch

from nomnomdata.engine.test_creds import update_creds


class NominodeMock(HTTMock):
    def __init__(self, task_parameters):
        super().__init__(self.api_match)
        self.logger = logging.getLogger("nomigen.nominode-mock")
        self.params = task_parameters
        self.secrets = self.params.pop("config", {})
        self.api_calls = []

    def __enter__(self):
        super().__enter__()
        self.old_environ = os.environ.copy()
        os.environ["execution_uuid"] = "TEST_UUID"
        os.environ["project_uuid"] = "TEST_PROJECT"
        os.environ["nomnom_api"] = "http://127.0.0.1:9090"
        os.environ["token"] = "token"
        return self

    def __exit__(self, type, value, traceback):
        super().__exit__(type, value, traceback)
        os.environ.clear()
        os.environ.update(self.old_environ)

    @urlmatch(netloc=r"(.*\.)?127.0.0.1:9090$")
    def api_match(self, url, request):
        match = re.match(r"/connection/(?P<uuid>.+)/update", url.path)
        if match:
            json_data = request.body
            loaded = json.loads(json_data)
            uuid = match.groupdict()["uuid"]
            self.params["config"][uuid] = json.loads(loaded["parameters"])
            update_creds(self.params["config"])
            self.logger.debug("Caught connections update. Test creds updated")
        elif url.path == "/execution/log/TEST_UUID":
            json_data = request.body
            # self.logger.debug("Caught log message {}".format(json_data))
            loaded = json.loads(json_data)
        elif url.path == "/task/TEST_UUID/update":
            json_data = request.body
            loaded = json.loads(json_data)
            self.logger.debug("Caught task update {}".format(pformat(loaded)))
        elif url.path == "/execution/update/TEST_UUID":
            json_data = request.body
            loaded = json.loads(json_data)
            self.logger.debug(
                "Caught execution progress update {}".format(pformat(loaded))
            )
        elif url.path == "/execution/decode/TEST_UUID":
            return json.dumps(self.secrets)
        elif url.path == "/task/TEST_UUID/parameters":
            return json.dumps({"result": "success"})
        elif url.path == "/execution/checkout/TEST_UUID":
            return json.dumps({"parameters": self.params, "task_uuid": "TASK_UUID"})
        else:
            self.logger.info(
                "Unknown api endpoint called {}, \n Body {}".format(
                    url.path, request.body.decode("utf-8")
                )
            )
        self.api_calls.append((url, request))
        return '{"you_logged":"test"}'


class BaseNomTest(unittest.TestCase):
    def setUp(self):
        self.params = {"config": {"nothing": "nothing"}, "action_name": "nothing"}
        self.api_calls = []
        self.logger = logging.getLogger("nomigen_tests." + self.__class__.__name__)
        self._old_environ = os.environ.copy()
        os.environ["execution_uuid"] = "TEST_UUID"
        os.environ["nomnom_api"] = "http://127.0.0.1:9090"
        os.environ["project_uuid"] = "TEST_PROJECT"
        os.environ["token"] = "token"

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self._old_environ)

    def run(self, result=None):
        with HTTMock(self.api_match):
            result = super().run(result)
        return result

    @urlmatch(netloc=r"(.*\.)?127.0.0.1:9090$")
    def api_match(self, url, request):
        match = re.match(r"/connection/(?P<uuid>.+)/update", url.path)
        if match:
            json_data = request.body
            loaded = json.loads(json_data)
            uuid = match.groupdict()["uuid"]
            self.params["config"][uuid] = json.loads(loaded["parameters"])
            update_creds(self.params["config"])
            self.logger.debug("Caught connections update. Test creds updated")
        elif url.path == "/execution/log/TEST_UUID":
            json_data = request.body
            self.logger.debug("Caught log message {}".format(json_data))
            loaded = json.loads(json_data)
        elif url.path == "/task/TEST_UUID/update":
            json_data = request.body
            loaded = json.loads(json_data)
            self.logger.debug("Caught task update {}".format(pformat(loaded)))
        elif url.path == "/execution/update/TEST_UUID":
            json_data = request.body
            loaded = json.loads(json_data)
            self.logger.debug(
                "Caught execution progress update {}".format(pformat(loaded))
            )
        elif url.path == "/execution/decode/TEST_UUID":
            return json.dumps(self.params["config"])
        elif url.path == "/task/TEST_UUID/parameters":
            return json.dumps({"result": "success"})
        elif url.path == "/execution/checkout/TEST_UUID":
            return json.dumps({"parameters": self.params, "task_uuid": "TASK_UUID"})
        else:
            self.logger.info("Unknown api endpoint called {}".format(url.path))
            self.logger.info(request.body.decode("utf-8"))
        self.api_calls.append((url, request))
        return '{"you_logged":"test"}'



import json
from dataclasses import dataclass
from enum import Enum as pyEnum
from typing import List

from nomnomdata.engine.components import ParameterType
from nomnomdata.engine.errors import ValidationError


class CodeDialectType(str, pyEnum):
    ABAP = "abap"
    ABC = "abc"
    ACTIONSCRIPT = "actionscript"
    ADA = "ada"
    APACHE_CONF = "apache_conf"
    APEX = "apex"
    APPLESCRIPT = "applescript"
    ASCIIDOC = "asciidoc"
    ASL = "asl"
    ASSEMBLY_X86 = "assembly_x86"
    AUTOHOTKEY = "autohotkey"
    BATCHFILE = "batchfile"
    BRO = "bro"
    C9SEARCH = "c9search"
    C_CPP = "c_cpp"
    CIRRU = "cirru"
    CLOJURE = "clojure"
    COBOL = "cobol"
    COFFEE = "coffee"
    COLDFUSION = "coldfusion"
    CSHARP = "csharp"
    CSOUND_DOCUMENT = "csound_document"
    CSOUND_ORCHESTRA = "csound_orchestra"
    CSOUND_SCORE = "csound_score"
    CSP = "csp"
    CSS = "css"
    CURLY = "curly"
    D = "d"
    DART = "dart"
    DIFF = "diff"
    DJANGO = "django"
    DOCKERFILE = "dockerfile"
    DOT = "dot"
    DROOLS = "drools"
    EDIFACT = "edifact"
    EIFFEL = "eiffel"
    EJS = "ejs"
    ELIXIR = "elixir"
    ELM = "elm"
    ERLANG = "erlang"
    FORTH = "forth"
    FORTRAN = "fortran"
    FSHARP = "fsharp"
    FSL = "fsl"
    FTL = "ftl"
    GCODE = "gcode"
    GHERKIN = "gherkin"
    GITIGNORE = "gitignore"
    GLSL = "glsl"
    GOBSTONES = "gobstones"
    GOLANG = "golang"
    GRAPHQLSCHEMA = "graphqlschema"
    GROOVY = "groovy"
    HAML = "haml"
    HANDLEBARS = "handlebars"
    HASKELL = "haskell"
    HASKELL_CABAL = "haskell_cabal"
    HAXE = "haxe"
    HJSON = "hjson"
    HTML = "html"
    HTML_ELIXIR = "html_elixir"
    HTML_RUBY = "html_ruby"
    INI = "ini"
    IO = "io"
    JACK = "jack"
    JADE = "jade"
    JAVA = "java"
    JAVASCRIPT = "javascript"
    JSON = "json"
    JSONIQ = "jsoniq"
    JSP = "jsp"
    JSSM = "jssm"
    JSX = "jsx"
    JULIA = "julia"
    KOTLIN = "kotlin"
    LATEX = "latex"
    LESS = "less"
    LIQUID = "liquid"
    LISP = "lisp"
    LIVESCRIPT = "livescript"
    LOGIQL = "logiql"
    LOGTALK = "logtalk"
    LSL = "lsl"
    LUA = "lua"
    LUAPAGE = "luapage"
    LUCENE = "lucene"
    MAKEFILE = "makefile"
    MARKDOWN = "markdown"
    MASK = "mask"
    MATLAB = "matlab"
    MAZE = "maze"
    MEL = "mel"
    MIXAL = "mixal"
    MUSHCODE = "mushcode"
    NIX = "nix"
    NSIS = "nsis"
    OBJECTIVEC = "objectivec"
    OCAML = "ocaml"
    PASCAL = "pascal"
    PERL = "perl"
    PERL6 = "perl6"
    PHP = "php"
    PHP_LARAVEL_BLADE = "php_laravel_blade"
    PIG = "pig"
    PLAIN_TEXT = "plain_text"
    POWERSHELL = "powershell"
    PRAAT = "praat"
    PROLOG = "prolog"
    PROPERTIES = "properties"
    PROTOBUF = "protobuf"
    PUPPET = "puppet"
    PYTHON = "python"
    R = "r"
    RAZOR = "razor"
    RDOC = "rdoc"
    RED = "red"
    RHTML = "rhtml"
    RST = "rst"
    RUBY = "ruby"
    RUST = "rust"
    SASS = "sass"
    SCAD = "scad"
    SCALA = "scala"
    SCHEME = "scheme"
    SCSS = "scss"
    SH = "sh"
    SJS = "sjs"
    SLIM = "slim"
    SMARTY = "smarty"
    SNIPPETS = "snippets"
    SOY_TEMPLATE = "soy_template"
    SPACE = "space"
    SPARQL = "sparql"
    STYLUS = "stylus"
    SVG = "svg"
    SWIFT = "swift"
    TCL = "tcl"
    TERRAFORM = "terraform"
    TEX = "tex"
    TEXT = "text"
    TEXTILE = "textile"
    TOML = "toml"
    TSX = "tsx"
    TURTLE = "turtle"
    TWIG = "twig"
    TYPESCRIPT = "typescript"
    VALA = "vala"
    VBSCRIPT = "vbscript"
    VELOCITY = "velocity"
    VERILOG = "verilog"
    VHDL = "vhdl"
    VISUALFORCE = "visualforce"
    WOLLOK = "wollok"
    XML = "xml"
    XQUERY = "xquery"
    YAML = "yaml"

    def __str__(self):
        return self.value


@dataclass
class Code(ParameterType):
    type = "code"
    dialect: CodeDialectType


@dataclass
class JSON(ParameterType):
    type = "json"

    def validator(self, val):
        try:
            json.loads(val)
        except json.JSONDecodeError:
            raise ValidationError("Not valid JSON")


@dataclass
class SQLDialectType(str, pyEnum):
    PGSQL = "pgsql"
    MYSQL = "mysql"
    REDSHIFT = "redshift"
    SQL = "sql"
    SQLSERVER = "sqlserver"


@dataclass
class SQL(ParameterType):
    type = "sql"
    dialect: SQLDialectType = SQLDialectType.SQL


@dataclass
class MetaDataTable(ParameterType):
    type = "metadata_table"


@dataclass
class Boolean(ParameterType):
    type = "boolean"


@dataclass
class Text(ParameterType):
    type = "text"


@dataclass
class Int(ParameterType):
    type = "int"
    max: int = None
    min: int = None

    def validate(self, val):
        if not isinstance(val, int):
            raise ValidationError(f"{type(val)} is not expected Integer type")
        if self.min and val < self.min:
            raise ValidationError(f"{val} is smaller than specified minimum [{self.min}]")
        if self.max and val > self.max:
            raise ValidationError(f"{val} is larger than specified maximum [{self.max}]")


@dataclass
class String(ParameterType):
    type = "string"


@dataclass
class Enum(ParameterType):
    type = "enum"
    choices: List[str]

    def validate(self, val):
        if not isinstance(val, str):
            raise ValidationError(f"{type(val)} is not a string")
        elif val not in self.choices:
            raise ValidationError(f"{val} is not a valid choice")


class Connection(ParameterType):
    type = "connection"



import json
import os.path

if "NOMIGEN_TEST_CREDENTIALS" in os.environ:
    creds_dir = os.environ["NOMIGEN_TEST_CREDENTIALS"]
else:
    creds_dir = "/nomnom/test_config"

if "CI" in os.environ:
    creds_path = os.environ["ENGINE_TEST_CREDS"]
else:
    creds_path = os.path.join(creds_dir, "nomigen_test_credentials.json")
if os.path.exists(creds_path):
    credentials = json.load(open(creds_path))
else:
    credentials = {}


def update_creds(new_creds):
    credentials = new_creds
    with open(creds_path, "w") as outfile:
        outfile.write(json.dumps(credentials, indent=4))


key_fp = os.path.join(creds_dir, "dummy_key.rsa")
private_key = None
if os.path.exists(key_fp):
    private_key = open(key_fp).read()



from json import JSONEncoder
from pathlib import PosixPath

from .base import Help, ParameterGroup, ParameterType


class ModelEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Help):
            return o.serialize()
        if isinstance(o, PosixPath):
            return str(o)
        if isinstance(o, ParameterGroup):
            return o.serialize()
        elif isinstance(o, ParameterType):
            return o.serialize()
        else:
            return super().default(o)



import json
import logging
from os import environ
from urllib.parse import urljoin, urlsplit, urlunsplit

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry


class NominodeClient:
    def __init__(self):
        self.logger = logging.getLogger("nomigen.nominode_api")
        self.token = environ["token"]
        self.execution_uuid = environ["execution_uuid"]
        self.project_uuid = environ["project_uuid"]
        # ensure we always have a trailing / on the api url, otherwise we mess up urljoin
        api_url = environ["nomnom_api"]
        self.nominode_url = api_url if api_url.endswith("/") else api_url + "/"
        self.session = requests.Session()
        self.session.headers.update({"token": self.token})
        # max value for each backoff is 2 minutes, 20 retries gets us about 30 minutes of retrying
        retries = Retry(total=20, backoff_factor=1, status_forcelist=[502, 503, 504, 404])
        self.session.mount(self.nominode_url, HTTPAdapter(max_retries=retries))

    def request(
        self,
        method,
        url_prefix=None,
        data=None,
        params=None,
        reg_data=None,
        headers=None,
        path_override=None,
    ):
        """
        Authenticated request to nominode.

        Makes an authenticated request to the nominode and returns a JSON blob.

        Parameters:
            method (string): HTTP Method to use (GET,POST,PUT,ect)
            url_prefix (string): Endpoint to hit eg execution/log
            data (dict): Payload for request, must be JSON serializable
            params (dict): URL Parameters, must be url encodable
            reg_data (string): Non JSON data to append to request. Cannot be used with data parameter
            headers (dict): Header dictionary
            path_override (string): Override string for the start of the nominode url (usually /api/v1)

        Returns:
            dict: JSON response data

        """
        response_data = self._get_response_data(
            method=method,
            endpoint_url=url_prefix,
            data=reg_data,
            headers=headers,
            json_data=data,
            params=params,
            path_override=path_override,
        )
        return response_data

    def _get_response_data(
        self,
        method,
        endpoint_url,
        json_data=None,
        params=None,
        data=None,
        headers=None,
        path_override=None,
    ):
        # endpoint urls should not start with a /
        endpoint_url = (
            endpoint_url if not endpoint_url.startswith("/") else endpoint_url.lstrip("/")
        )
        # this is to support overriding /api/1/ to /api/v2 if needed
        if path_override:
            split_url = list(urlsplit(self.nominode_url))
            split_url[2] = path_override
            base_url = urlunsplit(split_url)
            url = urljoin(base_url, endpoint_url)
        else:
            url = urljoin(self.nominode_url, endpoint_url)

        response = self.session.request(
            method, url, headers=headers, params=params, json=json_data, data=data
        )
        try:
            response.raise_for_status()
            data = response.json()
            if data:
                return data
            else:
                raise Exception("Empty json body response calling url: {}".format(url))
        except Exception:
            self.logger.exception(
                f"Error during {method}:{url} - {response.status_code} - response payload:\n{response.text}"
            )
            raise

    def update_progress(self, message=None, progress=None):
        """
        Update nominode with current task progress

        Parameters:
            message (string): Message for to attach to current task progress
            progress (int): Whole number representing the percentage complete of the task (0-100)
        Returns:
            dict: success/response data
        """
        # Called to periodically update the completion status of a given execution
        # Always sets to - '05': 'Executing: Running in docker container'
        if message is None and progress is None:
            raise Exception(
                "Message or Progress needs to be provided when updating execution status..."
            )
        data = {"status_code": "05", "progress": progress, "message": message}
        return self.request("put", "execution/update/%s" % self.execution_uuid, data=data)

    def update_connection(self, connection, connection_uuid):
        """
        Update a connection on the nominode

        Parameters:
            connection (dict): Dictionary representing the updated connection object
            connection_uuid (connection_uuid): UUID of the connection to be updated
        Returns:
            dict: success/response data
        """
        data = {"alias": connection["alias"], "parameters": json.dumps(connection)}
        return self.request("post", "connection/%s/update" % connection_uuid, data=data)

    def checkout_execution(self):
        """
        Fetch the task parameters. Should only be called once.

        Returns:
            dict: task parameters dictionary
        """
        return self.request("put", "execution/checkout/%s" % self.execution_uuid)

    def update_task_parameters(self, task_uuid, parameters):
        """
        Update the task parameters on the nominode

        Parameters:
            task_uuid (connection_uuid): UUID of this task
            parameters (dict): Dictionary representing what the new parameters will be. This overwrites all existing parameters.
        Returns:
            dict: success/response data
        """
        result = self.request(
            "put", "/task/{}/parameters".format(task_uuid), data=parameters
        )
        if "error" in result:
            raise Exception(result["error"])

    def update_result(self, result):
        """
        Update the task parameters on the nominode

        Parameters:
            result (dict): JSON data representing the result of this task. Currently only a json encoded Bokeh plot is supported.
        Returns:
            dict: success/response data
        """
        assert "result_type" in result
        assert "result" in result
        result = self.request(
            "put",
            f"projects/{self.project_uuid}/task_execution/{self.execution_uuid}/result",
            data=result,
            path_override="api/v2/",
        )
        if "error" in result:
            raise Exception(result["error"])
        else:
            return result

    def get_secrets(self):
        """
        Fetch the encoded connections associated with this task
        Returns:
            dict: decoded connections
        """
        x = self.request("get", "execution/decode/%s" % self.execution_uuid)
        if "error" in x:
            raise Exception(x["error"])
        else:
            return x

    def get_metadata_table(self, metadata_uuid, data_table_uuid):
        """
        Fetch a specific metadata table and data table from the nominode

        Parameters:
            metadata_uuid (string): UUID String of the metadata table.
            data_table_uuid (string): UUID String of the data table within the metadata.

        Returns:
            dict: success/response data
        """
        assert metadata_uuid, "metadata_uuid is required"
        assert data_table_uuid, "data_table_uuid is required"

        # Get the data_table details and column information
        url = "metadata/{metadata_uuid}/{data_table_uuid}"
        url = url.format(metadata_uuid=metadata_uuid, data_table_uuid=data_table_uuid)
        data_table = self.request("get", url)
        if "results" in data_table:
            # Grab first data table that matches.
            data_table = data_table["results"][0]
        else:
            raise Exception(
                "Error getting data table details for {}...".format(data_table_uuid)
            )

        return data_table



import json
import logging
import re
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from functools import wraps
from pathlib import Path
from typing import Callable, Dict, List

import click
import yaml

from nomnomdata.engine.api import NominodeClient
from nomnomdata.engine.errors import MissingParametersExection, ValidationError
from nomnomdata.engine.test import NominodeMock

from .base import Connection, Help, ParameterGroup
from .encoders import ModelEncoder

python_type = type
logger = logging.getLogger(__name__)


class RepoType(str, Enum):
    AWS = "aws"

    def __str__(self):
        return self.value


class ModelType(str, Enum):
    ENGINE = "engine"
    CONNECTION = "connection"
    SHARED_OBJECT = "shared_object"

    def __str__(self):
        return self.value


class Model:
    nnd_model_version = 2
    model_type: ModelType


@dataclass
class Action:
    name: str
    parameter_groups: List[ParameterGroup]
    display_name: str
    as_kwargs: bool
    description: str
    help: Help
    func: Callable

    @property
    def all_parameters(self):
        return {p.name: p for pg in self.parameter_groups.values() for p in pg.parameters}


class Icons:
    def __init__(self, one_x: Path, two_x: Path = None, three_x: Path = None):
        self.__dict__["1x"] = one_x
        if two_x:
            self.__dict__["2x"] = two_x
        if three_x:
            self.__dict__["3x"] = three_x


class Engine(Model):
    model_type = ModelType.ENGINE
    description: str
    alias: str
    image: str
    repo: str
    region: str
    repo_type: RepoType
    description: str
    categories: List[str]
    icons: Dict[str, Path]
    actions: Dict[str, Action]

    def __init__(
        self,
        uuid: str,
        alias: str,
        image: str,
        repo: str = "445607516549.dkr.ecr.us-east-1.amazonaws.com",
        region: str = "us-east-1",
        repo_type: RepoType = RepoType.AWS,
        description: str = "",
        categories: List[str] = None,
        icons: Icons = None,
    ):
        self.model = {
            "uuid": uuid,
            "alias": alias,
            "description": description,
            "location": {
                "image": image,
                "repo": repo,
                "region": region,
                "repo_type": str(repo_type),
            },
            "nnd_model_version": self.nnd_model_version,
            "categories": [{"name": val} for val in categories]
            if categories
            else [{"name": "General"}],
            "type": self.model_type.__str__(),
        }
        if icons:
            self.model["icons"] = icons.__dict__
        logger.debug(f"New Engine Registered '{uuid}'")
        self.actions = defaultdict(lambda: dict(parameters={}))
        super().__init__()

        @self.cli.command()
        def run():
            self.run()

        @self.cli.command()
        def dump_yaml():
            self.dump_yaml()

        @self.cli.command()
        @click.argument("action")
        @click.argument(
            "parameter_json_file", default="params.json", type=click.File("r")
        )
        def run_mock(action, parameter_json_file):
            self.run_mock(action, parameter_json_file)

    @staticmethod
    @click.group()
    def cli():
        pass

    def run(self):
        logger.info("Fetching task from nominode")
        client = NominodeClient()
        checkout = client.checkout_execution()
        params = checkout["parameters"]
        secrets = client.get_secrets()
        for secret_uuid in secrets:
            for pname, p in params.items():
                if isinstance(p, dict) and p.get("connection_uuid") == secret_uuid:
                    params[pname] = secrets[secret_uuid]
        logger.info(f"Action: {params['action_name']}")
        action = self.actions[params.pop("action_name")]
        kwargs = self.finalize_kwargs(action.all_parameters, params)
        logger.debug(f"Calling Action {action.name}")
        if action.as_kwargs:
            return action.func(**kwargs)
        else:
            return action.func(kwargs)

    def run_mock(self, action, parameter_json_file):
        params = json.load(parameter_json_file)
        action: Action = self.actions[action]
        kwargs = self.finalize_kwargs(action.all_parameters, params)
        connections = {
            pname: p
            for pname, p in action.all_parameters.items()
            if isinstance(p.type, Connection)
        }
        kwargs["config"] = {}
        for i, conn in enumerate(connections):
            kwargs[conn] = {"connection_uuid": str(i)}
            kwargs["config"][i] = params[conn]
        kwargs["action_name"] = action.name
        with NominodeMock(kwargs):
            logger.info("Nominode Mock in place")
            self.run()

    @staticmethod
    def check_missing(params, all_parameters, parent="."):
        missing_params = set([k for k, v in all_parameters.items() if v.required]) - set(
            params.keys()
        )
        if missing_params:
            for p in missing_params:
                logger.error(f"Missing Required Parameter {parent}:{p}")
            raise MissingParametersExection(f"{parent}:" + "+".join(missing_params))

    def finalize_kwargs(self, model_params, params, parent="."):
        kwargs = {}
        self.check_missing(params, model_params, parent)
        for keyword, val in params.items():
            parameter = model_params.get(keyword)
            if not parameter:
                logger.warning(f"\tUnknown parameter '{keyword}', discarding")
            elif isinstance(parameter.type, Connection):
                kwargs[keyword] = self.finalize_kwargs(
                    parameter.type.all_parameters, val, keyword
                )
            else:
                logger.debug(f"\tValidating {keyword}:'{val}' with {parameter.type}")
                try:
                    parameter.type.validate(val)
                except ValidationError:
                    logger.exception(f"\tParameter {parent}:{keyword} failed validation")
                    raise
                kwargs[keyword] = val
        return kwargs

    def dump_yaml(self):
        full_model = self.model.copy()
        actions = {}
        for name, action in self.actions.items():
            safe_name = re.sub("[^\w]", "", action.name.lower())
            actions[safe_name] = {
                "display_name": action.display_name or action.name.capitalize(),
                "description": action.description,
            }
            if action.help:
                actions[safe_name]["help"] = action.help
            actions[safe_name]["parameters"] = [
                p for p in action.parameter_groups.values()
            ]
        full_model["actions"] = actions
        json_dump = json.dumps(full_model, indent=4, cls=ModelEncoder)
        with open("model.yaml", "w") as f:
            f.write(yaml.dump(json.loads(json_dump), sort_keys=False))

    def main(self):
        self.cli.main()

    def action(self, name, help: Help = None, description="", as_kwargs=False):
        def action_dec(func):
            logger.debug(f"Action Registered '{name}'")
            self.actions[func.__name__] = Action(
                parameter_groups=func.parameter_groups,
                name=func.__name__,
                display_name=name,
                description=description,
                as_kwargs=as_kwargs,
                func=func,
                help=help,
            )

            @wraps(func)
            def call(*args, **kwargs):
                return self.__call__action__(func, *args, **kwargs)

            return call

        return action_dec

    def __call__action__(self, func, *args, **kwargs):
        with NominodeMock({}):
            action = self.actions[func.__name__]
            logger.debug(f"Action Called '{action.display_name}'")
            final_kwargs = self.finalize_kwargs(action.all_parameters, kwargs)
            if action.as_kwargs:
                return func(**final_kwargs)
            else:
                return func(final_kwargs)

    def parameter_group(
        self, parameter_group: ParameterGroup, name="", description="", alias=""
    ):
        def parameter_dec(func):
            logger.debug(f"\tParameter Group Registered '{name}'")
            params = getattr(func, "parameter_groups", {})
            parameter_group.name = name
            params[name] = parameter_group
            func.parameter_groups = params
            return func

        return parameter_dec



import collections
import json

from nomnomdata.engine import api
from nomnomdata.engine.logging import getLogger


class EngineParameters(collections.MutableMapping):
    """
        A Dict like collection representing the parameters the running task can access.
    """

    def __init__(self, parameters, task_uuid):
        self.nominode_client = api.NominodeClient()

        self.task_uuid = task_uuid
        if isinstance(parameters, str):
            self.__params = json.loads(parameters)
        else:
            self.__params = parameters
        self.__params["config"] = self.nominode_client.get_secrets()
        self.__connections = []
        for key, values in self.__params.items():
            if isinstance(values, dict) and "connection_uuid" in values:
                self.__connections.append(key)
                self.__params[key].update(
                    self.__params["config"][values["connection_uuid"]]
                )

    def __getitem__(self, key):
        return self.__params[key]

    def __setitem__(self, key, value):
        raise NotImplementedError(
            "Cannot set parameter values directly, use foo.commit(key, value)"
        )

    def commit(self, key, value, local_only=False):
        """
        Send the new parameter value to the nominode.

        Parameters:
            key (string): Parameter key.
            value (dict|list|tuple|bool|None): New value of the parameter. Must be JSON serializable or an exception will occur unless local_only is True.
            local_only (boolean): Set to True to only update the value of the parameter locally and not update the nominode stored version
        Returns:
            None
        """
        if key in self.__connections or key == "config":
            raise NotImplementedError("Updating connections is not currently supported")
        if not local_only:
            self.nominode_client.update_task_parameters(self.task_uuid, {key: value})
        self.__params[key] = value

    def __delitem__(self, key):
        raise ValueError("Cannot delete parameters")

    def __iter__(self):
        return iter(self.__params)

    def __len__(self):
        return len(self.__params)

    def __keytransform__(self, key):
        return key

    def __str__(self):
        return str(self.__params)


class Executable(object):
    """
        The key base class for any task. This must be inherited.
        WARNING : super().__init__() MUST be called before doing anything else in the task!.
        This sets up self.params,self.nominode_client and self.logger for use by the task.
        Extend this class with extra functions that are your code to be run.
        self.logger is special in that it logs all messages to the nominode, using other loggers means users will not see those messages.
    """

    def __init__(self, logger_name="nomigen"):
        self.nominode_client = api.NominodeClient()
        self.logger = getLogger(logger_name)
        self.checkout = self.nominode_client.checkout_execution()
        self.params = EngineParameters(
            self.checkout["parameters"], self.checkout["task_uuid"]
        )
        self.action = self.params.get("action_name", "no_action")
        self.task_uuid = self.checkout["task_uuid"]

    def get_action(self):
        return self.action

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        pass



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List


class ParameterType(metaclass=ABCMeta):
    @property
    @abstractmethod
    def type(self):
        pass

    def validate(self, val):
        pass

    def serialize(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def convert(self, val):
        return val


@dataclass
class Help:
    header_id: str = None
    md_path: Path = None

    def serialize(self):
        if self.header_id:
            return {"header_id": self.header_id}
        elif self.md_path:
            return {"file": str(self.md_path)}


class Parameter:
    type: ParameterType
    name: str
    help: Help
    display_name: str
    required: bool
    description: str
    default: object

    def __init__(
        self,
        type: ParameterType,
        name: str,
        display_name: str = "",
        help: Help = None,
        required: bool = False,
        description: str = "",
        default: object = None,
    ):
        self.type = type
        self.name = name
        self.help = help
        self.display_name = display_name or self.name.capitalize()
        self.required = required
        self.description = description
        self.default = default

    def serialize(self):
        result = {
            "name": self.name,
            "display_name": self.display_name,
            "description": self.description,
            "type": self.type.type,
            "required": self.required,
        }
        if self.help:
            result["help"] = self.help.serialize()
        if self.default:
            result["default"] = self.default
        result.update(self.type.serialize())
        return result

    def validate(self, val):
        return self.type.validate(val)

    def convert(self, val):
        return self.type.convert(val)

    def __str__(self):
        return f"{self.type.__class__.__name__} Parameter"

    def __repr__(self):
        return f"<{self.type.__class__.__name__} Parameter>"


class ParameterGroup:
    type = "group"
    parameters = List[Parameter]
    collapsed: bool
    shared_parameter_group_uuid: str
    description: str
    name: str
    display_name: str

    def __init__(
        self,
        *args: List[Parameter],
        display_name: str = "",
        name: str = "General",
        collapsed: bool = False,
        shared_parameter_group_uuid: str = "",
        description: str = "",
    ):
        self.parameters = args
        self.collapsed = collapsed
        self.shared_parameter_group_uuid = shared_parameter_group_uuid
        self.description = description
        self.name = name
        self.display_name = display_name

    def serialize(self):
        result = {
            "name": self.name,
            "display_name": self.display_name or self.name.capitalize(),
            "description": self.description,
            "type": self.type,
            "collapsed": self.collapsed,
            "parameters": [p.serialize() for p in self.parameters],
        }
        if self.shared_parameter_group_uuid:
            result["shared_parameter_group_uuid"] = (self.shared_parameter_group_uuid,)
        return result


class Connection(ParameterType, metaclass=ABCMeta):
    type = "connection"

    @property
    def all_parameters(self):
        return {p.name: p for pg in self.parameter_groups.values() for p in pg.parameters}

    @property
    @abstractmethod
    def connection_type_uuid(self) -> str:
        pass

    @property
    @abstractmethod
    def parameter_groups(self) -> Dict[str, ParameterGroup]:
        pass



import logging
import os
import traceback
from concurrent.futures import ThreadPoolExecutor

from nomnomdata.engine.api import NominodeClient


class NominodeLogHandler(logging.Handler):
    # Send log lines in a separate thread, log lines are 'non critical' so if we get a timeout we don't care too much
    # Using a thread pool to limit the number of threads spawned

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.threadPool = ThreadPoolExecutor(4)
        self.nominode = NominodeClient()

    def emit(self, record):
        self.threadPool.submit(self._emit, record)

    def _emit(self, record):
        to_send = record.__dict__.copy()
        to_send["execution_uuid"] = os.environ["execution_uuid"]
        to_send["log_version"] = "0.1.0"
        if to_send["exc_info"]:
            to_send["exception_lines"] = traceback.format_exception(*to_send["exc_info"])
            del to_send["exc_info"]
        to_send["msg"] = str(to_send["msg"])
        self.nominode.request(
            "put", "execution/log/%s" % os.environ["execution_uuid"], data=to_send
        )


def getLogger(name):
    logger = logging.getLogger(name)
    if os.environ.get("token"):
        logger.setLevel(os.environ.get("log_level") or "INFO")
        handler = NominodeLogHandler()
        logger.addHandler(handler)
    return logger



import importlib.util
import json
import os
import subprocess
import sys
from os import path

import click
import pytest

from nomnomdata.engine.core import Executable
from nomnomdata.engine.test import NominodeMock


def clear_pycache():
    subprocess.check_call("find . -name '*.pyc' -delete", shell=True)


@click.command(name="run_tests")
@click.option(
    "-t", "--test", help="Run a specific test name (equivalent to pytest -k option)"
)
@click.option(
    "-c",
    "--creds_dir",
    default="/nomnom/test_config",
    show_default=True,
    envvar="NOMIGEN_TEST_CREDENTIALS",
    help="Credentials folder, will try environment variable NOMIGEN_TEST_CREDENTIALS, before falling back to default",
)
@click.pass_context
def run_tests(ctx, creds_dir=None, test=None):
    """Run engine tests with pytest, checks NOMIGEN_TEST_CREDENTIALS environment variable for creds location, otherwise uses /nomnom/test_config"""
    clear_pycache()
    old_environ = os.environ.copy()
    try:
        os.environ["log_level"] = ctx.obj["LOG_LEVEL"]
        xml_dest = os.path.join(creds_dir, "test-reports", "nomtest.xml")
        pytestargs = ["-s", "-v", "--cache-clear", "--junitxml={}".format(xml_dest)]
        if test:
            pytestargs.extend(["-k", test])
        pytestargs.extend(["--cov-fail-under=80", "--cov=pkg"])
        result = pytest.main(pytestargs)
    finally:
        os.environ.clear()
        os.environ.update(old_environ)
    sys.exit(result)


@click.command()
@click.option(
    "-p",
    "--parameter_file",
    default="params.json",
    show_default=True,
    help="Parameter JSON file to use",
)
@click.argument("action_name")
@click.argument("executable", default="executable.py")
@click.pass_context
def run_mock(ctx, action_name, executable, parameter_file):
    """Run the target executable, expected to be a python file that contains a subclass of core.Executable"""
    clear_pycache()
    old_environ = os.environ.copy()
    try:
        with open(parameter_file, "r") as pfile:
            params = json.load(pfile)
        os.environ["log_level"] = ctx.obj["LOG_LEVEL"]
        spec = importlib.util.spec_from_file_location(
            "executable", path.abspath(executable)
        )
        module = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = module
        spec.loader.exec_module(module)
        executables = Executable.__subclasses__()
        with NominodeMock(params):
            executable = executables[0]
            with executable() as r:
                to_run = getattr(r, action_name)
                to_run()
    finally:
        os.environ.clear()
        os.environ.update(old_environ)



from dataclasses import dataclass

from nomnomdata.engine.components import Connection, Parameter, ParameterGroup
from nomnomdata.engine.parameters import Int, String


@dataclass
class AWSTokenConnection(Connection):
    parameter_groups = {
        "Auth": ParameterGroup(
            Parameter(type=String(), name="token", required=True),
            Parameter(type=Int(), name="number", required=True),
            name="auth",
            description="Auth parameters",
        )
    }
    connection_type_uuid: str = "AWS5D-TO99M"


@dataclass
class FTPConnection(Connection):
    parameter_groups = {
        "Auth": ParameterGroup(
            Parameter(type=String(), name="hostname"),
            name="auth",
            description="Auth parameters",
        )
    }
    connection_type_uuid: str = "FTP92-TS0BZ"



import importlib.util
import sys
from os import path

import click

from nomnomdata.engine import logging as nomlogging
from nomnomdata.engine.api import NominodeClient
from nomnomdata.engine.core import Executable


@click.command()
@click.argument("executable", default="executable.py")
def run(executable):
    """ Run the target executable, expected to be a python file that contains a subclass of core.Executable"""
    nominode = NominodeClient()
    logger = nomlogging.getLogger("nomigen.run")
    spec = importlib.util.spec_from_file_location("executable", path.abspath(executable))
    module = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = module
    spec.loader.exec_module(module)
    executables = Executable.__subclasses__()
    if len(executables) != 1:
        logger.error(
            "Executable can only be subclassed once, current subclasses {}".format(
                executables
            )
        )
        sys.exit(1)
    else:
        executable = executables[0]
        with executable() as r:
            to_run = getattr(r, r.get_action())
            to_run()
        nominode.update_progress(progress=100)



import click

from . import __version__
from .run import run
from .run_tests import run_mock, run_tests


@click.group(name="engine")
@click.version_option(version=__version__, prog_name="nomnomdata-engine")
def cli():
    """NomNomData Engine CLI, used for the execution of engines"""


@cli.command()
def test():
    """Run unittests for the nomnomdata-engine module"""
    import pytest

    pytest.main(["--pyargs", "nomnomdata.engine"])


cli.add_command(run)
cli.add_command(run_tests)
cli.add_command(run_mock)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nomnomdata',
 'nomnomdata.engine',
 'nomnomdata.engine.components',
 'nomnomdata.engine.connections',
 'nomnomdata.engine.errors',
 'nomnomdata.engine.parameters',
 'nomnomdata.engine.tests']

package_data = \
{'': ['*']}

install_requires = \
['dunamai>=1.1.0,<2.0.0',
 'httmock>=1.3.0,<2.0.0',
 'nomnomdata-cli>=0.1.0,<0.2.0',
 'pytest-cov>=2.8.1,<3.0.0',
 'pytest>=5.4.1,<6.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.23.0,<3.0.0']

entry_points = \
{'nomnomdata.cli_plugins': ['engine = nomnomdata.engine.cli:cli']}

setup_kwargs = {
    'name': 'nomnomdata-engine',
    'version': '0.1.3.post12.dev0',
    'description': 'Package containing tooling for developing nominode engines',
    'long_description': 'Package for developing nominode engines\n',
    'author': 'Nom Nom Data Inc',
    'author_email': 'info@nomnomdata.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/nomnomdata/tools/nomnomdata-engine',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
